var soap = require('soap');

var wsdlName = "BsaBase", wsdl = __dirname + '/data/wsdl/' + wsdlName + '.wsdl';

var settings = {
	helloTimeout: 2000,
	soapTimeout: 5000
};

/**
 * Validates URL address.
 * 
 * @private
 * @author Billon sp. z o.o.
 * @param {string} str URL address
 * @return {boolean} True if the address is correct, otherwise false.
 */
var ValidURL = function(str) {
	var pattern = new RegExp('^(https?:\\/\\/)?'+ // protocol
	'((([a-z\\d]([a-z\\d-]*[a-z\\d])*)\\.)+[a-z]{2,}|'+ // domain name
	'((\\d{1,3}\\.){3}\\d{1,3}))'+ // OR ip (v4) address
	'(\\:\\d+)?(\\/[-a-z\\d%_.~+]*)*'+ // port and path
	'(\\?[;&a-z\\d%_.~+=-]*)?'+ // query string
	'(\\#[-a-z\\d_]*)?$','i'); // fragment locator
	if(!pattern.test(str)) {
		return false;
	} else {
		return true;
	}
};

var errors = {
	ERR_EMPTY_REQUEST_BODY: {
		"err": 'ERR_EMPTY_REQUEST_BODY',
		"description": 'Request body must not be a valid JSON.'
	},
	ERR_MISSING_WUI_DATA: {
		"err": 'ERR_MISSING_WUI_DATA',
		"description": 'Missing parameter of CNode in request body.'
	},
	ERR_MISSING_WUI_ADDRESS: {
		"err": 'ERR_MISSING_WUI_ADDRESS',
		"description": 'Missing WUI.Address parameter in request body.'
	},
	ERR_MISSING_WUI_NAME: {
		"err": 'ERR_MISSING_WUI_NAME',
		"description": 'Missing WUI.Name parameter in request body.'
	},
	ERR_INVALID_WUI_ADDRESS_DATA_TYPE: {
		"err": 'ERR_INVALID_WUI_ADDRESS_DATA_TYPE',
		"description": 'cNode.Address parameter must be a string.'
	},
	ERR_INVALID_WUI_NAME_DATA_TYPE: {
		"err": 'ERR_INVALID_WUI_NAME_DATA_TYPE',
		"description": 'cNode.Name parameter must be a string.'
	},
	ERR_INVALID_WUI_ADDRESS: {
		"err": 'ERR_INVALID_WUI_ADDRESS',
		"description": 'cNode.Address is not a valid URL.'
	},
	ERR_INVALID_PARAMETERS: {
		"err": 'ERR_INVALID_PARAMETERS',
		"description": 'Invalid request body parameters for that request.'
	},
	ERR_WUI_NOT_RESPONDING: {
		"err": 'ERR_WUI_NOT_RESPONDING',
		"description": 'CNode is not responding.'
	},
	ERR_COULD_NOT_HELLO:  {
		"err": 'ERR_COULD_NOT_HELLO',
		"description": 'Hello request failed.'
	},
	ERR_WRONG_WUI_USERNAME:  {
		"err": 'ERR_WRONG_WUI_USERNAME',
		"description": 'Requested CNode has different username then specified.'
	},
	ERR_REQUEST_BODY_IS_NOT_VALID_JSON: {
		"err": 'ERR_REQUEST_BODY_IS_NOT_VALID_JSON',
		"description": 'Request body must be a valid JSON.'
	},
	ERR_COULD_NOT_GET_TASK_STATUS: {
		"err": 'ERR_COULD_NOT_GET_TASK_STATUS',
		"description": 'Could not get task status.'
	},
	ERR_WRONG_WUI_RESPONSE: {
		"err": 'ERR_WRONG_WUI_RESPONSE',
		"description": 'Received wrong response from CNode.'
	},
	ERR_INVALID_RESULT: {
        "err": 'ERR_INVALID_RESULT',
        "description": 'CNode returned empty response.'
    }
};

/**
 * Validates Corporate Node data in a object and returns error.
 * 
 * @private
 * @author Billon sp. z o.o.
 * @param {object} request Corporate Node data and parameters of the SOAP message.
 * @return {(boolean|object))} False or object.
 */
var invalid = function (request) {
	var err = false;

	if (!request) {
		err = errors.ERR_EMPTY_REQUEST_BODY;
	} else if (!request.cNode) {
		if (request.WUI) {
			request.cNode = request.WUI;
		}
		else {
			err = errors.ERR_MISSING_WUI_DATA;
		}
	} else if (typeof request.cNode == "object") {
		if (typeof request.cNode.Address != "string") {
			err = errors.ERR_INVALID_WUI_ADDRESS_DATA_TYPE;
		} else if (typeof request.cNode.Name != "string") {
			err = errors.ERR_INVALID_WUI_NAME_DATA_TYPE;
		} else if (!ValidURL(request.cNode.Address)) {
			err = errors.ERR_INVALID_WUI_ADDRESS;
		}
	}

	return err;
};

/**
 * Runs a function, which is the fourth argument, or sends a request to service 'billon.me'.
 * @since 2.2
 * 
 * @private
 * @author Billon sp. z o.o.
 * @param {boolean} directly Whether the function that is the fourth argument is to called instead of communication with the service 'billon.me'.
 * @param {(function|undefined|null)} onBillonMeRequest A function called after receiving a response from the service 'billon.me'
 * @param {object} billonMeRequest URL A object that is to be sent to the service 'billon.me'.
 * @param {(undefined|null|function)} soapRequest The function called if first argument is true.
 */
function runRequest(directly, onBillonMeRequest, billonMeRequest, soapRequest) {
	if (directly) {
		return soapRequest();
	}
	else {
		var request = require('request');
		request.post({"rejectUnauthorized": false, url: 'https://billon.me/api/v1', form: billonMeRequest}, function (error, response, body) {
			if (error) {
				request.post({"rejectUnauthorized": false, url: 'https://billon.me:5002/api/v1', form: billonMeRequest}, 
							 function (error, response, body) {
					if (error) {
						onBillonMeRequest(error, body);
					}
					else {
						return onBillonMeRequest(error, JSON.parse(body));
					}
				});
			}
			else {
				try {
					return onBillonMeRequest(error, JSON.parse(body));
				}
				catch (e) {
					return onBillonMeRequest(error, body);
				}
			}
		});
	}
}

/**
 * Sends SOAP request 'hello' to Corporate Node and calls a function with an error.
 * 
 * @private
 * @author Billon sp. z o.o.
 * @param {object} wuiName A token from service 'billon.me' or an username of Corporate Node.
 * @param {function} cb The function to be called.
 * @param {object} options an object with positive parameter 'billon.me' or options of SOAP client.
 * @param {(undefined|null|Client)} client SOAP client for direct communication with Corporate Node.
 */
var helloCheck = function(wuiName, cb, options, client) {
	try {
		var directly = !options || typeof options != 'object' || !options['billon.me'];
		runRequest(directly, function(err, result) {
			if (typeof result !== "object") {
				var error = errors.ERR_INVALID_RESULT;
				error.native_err = result;
				return cb(error);
			}
			if (err) {
				var error = errors.ERR_COULD_NOT_HELLO;
				error.native_err = err;
				cb(error);
			} else if (typeof result !== "object" || result.constructor === Array) {
				var error = errors.ERR_WRONG_WUI_RESPONSE;
				error.native_err = result;
				cb(error);
			} else {
				cb();
			}
			SoapService.onRequestSend('hello', null, result, err);
		}, {token: wuiName, request: 'hello'}, function() {
			client.hello(null, function(err, result) {
				var resultIsNotObject = typeof result !== "object" || typeof result.return !== "object";
				SoapService.onRequestSend('hello', null, resultIsNotObject ? result : result.return, err);
				if (resultIsNotObject) {
					var error = errors.ERR_INVALID_RESULT;
					error.native_err = result;
					return cb(error);
				}

				var fixedUsername = result.return.username.replace(/_/g, "-").replace(/\./g, "");
				var fixedWuiName = wuiName.replace(/_/g, "-").replace(/\./g, "");
				if (err) {
					var error = errors.ERR_COULD_NOT_HELLO;
					error.native_err = err;
					cb(error);
				} else if (resultIsNotObject || typeof result.return.username !== "string") {
					var error = errors.ERR_WRONG_WUI_RESPONSE;
					error.native_err = result;
					cb(error);
				} else if (fixedUsername != fixedWuiName) {
					cb(errors.ERR_WRONG_WUI_USERNAME);
				} else {
					cb();
				}
			}, options);
		});
	} catch (err) {
		var error = errors.ERR_WUI_NOT_RESPONDING;
		error.native_err = err;
		cb(error);
	}
};

var SoapService = {
	
	/**
     * Runs after receiving a response of Corporate Node or a service 'billon.me'
	 * @since 2.1
	 * 
	 * @param  {string} requestName Name of the SOAP message.
	 * @param  {object|null} parameters Parameters of the SOAP message.
	 * @param  {object|null|undefined} requestResponse Corporate Node response
	 * @param  {object|null|undefined} error Errors when sending SOAP request
	 */
	onRequestSend: function (requestName, parameters, requestResponse, error) {},
	
	/**
     * Sends SOAP request to Corporate Node and outputs response.
	 * 
	 * @param  {string} method Name of the SOAP message.
	 * @param  {object} requestBody Corporate Node data and parameters of the SOAP message.
	 * @param  {Response} res Sails.js response object.
	 */
	execute: function (method, requestBody, res) {
		this.call[method](requestBody, function(data){
			res.json(data);
		});
	},

	/**
     * Sends SOAP request to Corporate Node.
	 * 
	 * @param  {string} requestName Name of the SOAP message.
	 * @param  {object} requestBody Corporate Node data and parameters of the SOAP message.
	 * @param  {function} cb A function that is to be called with a response or an error after receiving the response.
	 */
	request: function (requestName, requestBody, cb) {
		err = invalid(requestBody);
		if (err) {
			cb(err);
			return;
		}

		if (!requestBody.params) {
			requestBody.params = null;
		}
		var directly = typeof requestBody.cNode == 'object';
		function runHelloCheck(client) {
			helloCheck(directly ? requestBody.cNode.Name : requestBody.cNode, function(err) {
				if (err) {
					cb(err);
					return;
				}
	
				try {
					runRequest(directly, function(err, result) {
						SoapService.onRequestSend(requestName, requestBody.params, result, err);
						if (err)
						{
							var error = errors.ERR_INVALID_PARAMETERS;
							error.native_err = err;
							cb(error);
						} else {
							if (typeof result.data === "undefined" && typeof requestBody.data !== "undefined") {
								result.data = requestBody.data;
							}
							cb(result);
						}
					}, {token: requestBody.cNode, request: requestName, params: requestBody.params}, function() {
						client[requestName](requestBody.params, function(err, result) {
							SoapService.onRequestSend(requestName, requestBody.params, typeof result !== "object" ? result : result.return, err);
							if (err) {
								var error = errors.ERR_INVALID_PARAMETERS;
								error.native_err = err;
								cb(error);
							} else {
								var response = result.return;
								if (typeof response.data === "undefined" && typeof requestBody.data !== "undefined") {
									response.data = requestBody.data;
								}
								cb(response);
							}
						}, {timeout: settings.soapTimeout});
					});
				} catch (error) {
					var errorData = errors.ERR_INVALID_PARAMETERS;
					errorData.native_err = error;
					cb(errorData);
				}
			}, directly ? {timeout: settings.soapTimeout} : {'billon.me': true}, client);
		}
		if (directly) {
			soap.createClient(wsdl, function(err, client) {
				if (err) {
					cb(err);
				}
				else {
					client.setEndpoint(requestBody.cNode.Address);
					runHelloCheck(client);
				}
			});
		}
		else {
			try {
				var time = Date.now();
				runRequest(false, function(err, result) {
					SoapService.onRequestSend(requestName, requestBody.params, result, err);
					if (err)
					{
						runHelloCheck();
					} else {
						if (result && typeof result != "undefined" && result.constructor !== Array) {
							cb(result);
						}
						else {
							if (Date.now() - time < 1000) {
								runHelloCheck();
							}
							else {
								var error = errors.ERR_WRONG_WUI_RESPONSE;
								error.native_err = result;
								cb(error);
							}
						}
					}
				}, {token: requestBody.cNode, request: requestName, params: requestBody.params}, function() {});
			} catch (error) {
				runHelloCheck();
			}
		}
	}
};

SoapService.call = {
	SoapService: SoapService
};
soap.createClient(wsdl, function(err, cli) {
	if (err) {
		console.log(err);
		return;
	}
	for (var i in cli[wsdlName][wsdlName]) {
		SoapService.call[i] = new Function("requestBody", "cb", "this.SoapService.request('"+i+"', requestBody, function(data){cb(data);});");
	}

});

module.exports = SoapService;
